/*
 * Copyright (c) 1999 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/types.h>
#include <oskit/io/bufio.h>
#include <oskit/net/ether.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/osenv.h>
#include <oskit/dev/net.h>
#include <oskit/dev/ethernet.h>
#include <oskit/startup.h>
#include <stdio.h>
#include <stdlib.h>
#include <oskit/debug.h>
#include "netio_guts.h"

#define PRINTF(fmt, args... )  printf(__FUNCTION__  ":" fmt , ## args)
#if 1
#define DPRINTF(fmt, args... ) printf(__FUNCTION__  ":" fmt , ## args)
#else
#define DPRINTF(fmt, args... )
#endif

/*
 * Private data structure for each opened device.
 */
struct etherdev {
	oskit_etherdev_t	*dev;
	oskit_netio_t		*send_nio;
	oskit_netio_t		*recv_nio;
	oskit_devinfo_t		info;
	unsigned char		haddr[OSKIT_ETHERDEV_ADDR_SIZE];
	void			*obj;	/* C++ object to pass back */
	char			send_inuse;
	char			recv_inuse;
};
#define MAX_EDEVS	4
static struct etherdev	*myedevs[MAX_EDEVS];
static int		netio_init_guts(int ifnum);
static unsigned char	bcastaddr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/*
 * The network callback. This will suck out a data pointer, and then
 * pass that up to the c++ class callback.
 */
static oskit_error_t
net_receive(void *data, oskit_bufio_t *b, oskit_size_t pkt_size)
{
	struct etherdev		*edev = (struct etherdev *)data;
	oskit_error_t		err;
	void			*frame;
	struct ether_header	*eth;
	
	err = oskit_bufio_map(b, (void **)&frame, 0, pkt_size);
	assert(err == 0);
	eth = (struct ether_header *)frame;

	if (memcmp(eth->ether_dhost, &edev->haddr, sizeof(edev->haddr)) != 0
	    && memcmp(eth->ether_dhost, bcastaddr, sizeof bcastaddr) != 0) 
		return 0;

	/*
	 * Pass bufio up to C++ NetIO class. It will handle incrementing
	 * the reference count if it decides to hang onto it.
	 */
	if (edev->recv_inuse)
		click_FromNetio_in(edev->obj, b, (int) pkt_size);

	return 0;
}

int
oskit_create_fromnetio(void *obj, int ifnum, void **priv)
{
	struct etherdev		*edev;

	if (ifnum < 0 || ifnum >= MAX_EDEVS) {
		PRINTF("Bad ifnum:%d. Max ifnum is %d\n", ifnum, MAX_EDEVS);
		return -1;
	}
	if ((edev = myedevs[ifnum]) == NULL) {
		if (netio_init_guts(ifnum))
			return -1;
		edev = myedevs[ifnum];
	}

	/*
	 * Not great, but lets see how we want to use this before we
	 * get fancy about reference counting.
	 */
	if (edev->recv_inuse) {
		PRINTF("ifnum:%d already in use.\n", ifnum);
		return -1;
	}

	/*
	 * Stash the object pointer for callback.
	 */
	edev->obj        = obj;
	edev->recv_inuse = 1;
	*priv            = (void *) edev;

	return 0;
}

int
oskit_create_tonetio(void *obj, int ifnum, void **priv)
{
	struct etherdev		*edev;

	if (ifnum < 0 || ifnum >= MAX_EDEVS) {
		PRINTF("Bad ifnum:%d. Max ifnum is %d\n", ifnum, MAX_EDEVS);
		return -1;
	}
	if ((edev = myedevs[ifnum]) == NULL) {
		if (netio_init_guts(ifnum))
			return -1;
		edev = myedevs[ifnum];
	}

	/*
	 * Not great, but lets see how we want to use this before we
	 * get fancy about reference counting.
	 */
	if (edev->send_inuse) {
		PRINTF("ifnum:%d already in use.\n", ifnum);
		return -1;
	}

	edev->send_inuse = 1;
	*priv            = (void *) edev;

	return 0;
}

int
oskit_destroy_fromnetio(int ifnum, void *priv)
{
	struct etherdev		*edev;

	if (ifnum < 0 || ifnum >= MAX_EDEVS) {
		PRINTF("Bad ifnum:%d. Max ifnum is %d\n", ifnum, MAX_EDEVS);
		return -1;
	}
	if ((edev = myedevs[ifnum]) == NULL) {
		PRINTF("Bad ifnum:%d. Interface never used\n", ifnum);
		return -1;
	}

	if (edev->recv_inuse == 0) {
		PRINTF("Bad ifnum:%d. Interface never used\n", ifnum);
		return -1;
	}

	edev->recv_inuse = 0;
	if (edev->recv_inuse || edev->send_inuse)
		return 0;

	DPRINTF("Destroying NetIO %d ...\n", ifnum);

	oskit_etherdev_release(edev->dev);
	oskit_netio_release(edev->send_nio);
	oskit_netio_release(edev->recv_nio);
	free(edev);
	myedevs[ifnum] = 0;

	return 0;
}

oskit_destroy_tonetio(int ifnum, void *priv)
{
	struct etherdev		*edev;

	if (ifnum < 0 || ifnum >= MAX_EDEVS) {
		PRINTF("Bad ifnum:%d. Max ifnum is %d\n", ifnum, MAX_EDEVS);
		return -1;
	}
	if ((edev = myedevs[ifnum]) == NULL) {
		PRINTF("Bad ifnum:%d. Interface never used\n", ifnum);
		return -1;
	}

	if (edev->send_inuse == 0) {
		PRINTF("Bad ifnum:%d. Interface never used\n", ifnum);
		return -1;
	}

	edev->send_inuse = 0;
	if (edev->recv_inuse || edev->send_inuse)
		return 0;

	DPRINTF("Destroying NetIO %d ...\n", ifnum);

	oskit_etherdev_release(edev->dev);
	oskit_netio_release(edev->send_nio);
	oskit_netio_release(edev->recv_nio);
	free(edev);
	myedevs[ifnum] = 0;

	return 0;
}

/*
 * Click hands us a bufio to send. Push it out.
 */
int
oskit_sendto_netio_withbufio(void *obj, oskit_bufio_t *b, int len)
{
	struct etherdev	*edev = (struct etherdev *) obj;
	
	oskit_netio_push(edev->send_nio, b, len);
	
	return 0;
}

/*
 * Initialize the netio interface for a specific (integer) interface.
 * We pass back (through the priv pointer) a pointer to the etherdev.
 */
static int
netio_init_guts(int ifnum)
{
	oskit_error_t		err;
	oskit_etherdev_t	**etherdev;
	int			i, ndev;
	struct etherdev		*edev;

	/*
	 * Find all the Ethernet device nodes.
	 */
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&etherdev);
	if (ndev <= 0) {
		PRINTF("no Ethernet adaptors found!");
		goto bad;
	}
	if (ifnum < 0 || ifnum >= ndev || ifnum >= MAX_EDEVS) {
		PRINTF("Bad ifnum:%d. Max ifnum is %d\n", ifnum, ndev);
		goto bad;
	}

	if ((edev = (struct etherdev *) malloc(sizeof(*edev))) == NULL) {
		PRINTF("Out of memory allocating etherdev\n");
		goto bad;
	}
	memset(edev, 0, sizeof(*edev));

	err = oskit_etherdev_getinfo(etherdev[ifnum], &(edev->info));
	if (err) {
		PRINTF("Error getting info from ethercard %d", ifnum);
		goto bad;
	}
	
	oskit_etherdev_getaddr(etherdev[ifnum], edev->haddr);

	/*
	 * Show information about this adaptor
	 */
	DPRINTF("%-16s%-40s\n", edev->info.name,
		edev->info.description ? edev->info.description : "");

	edev->dev      = etherdev[ifnum];
	edev->recv_nio = oskit_netio_create(net_receive, edev);
	if (edev->recv_nio == NULL) {
		PRINTF("unable to create recv_nio");
		goto bad;
	}
	
	err = oskit_etherdev_open(etherdev[ifnum],
				  0, edev->recv_nio, &(edev->send_nio));

	if (err) {
		PRINTF("Error opening ethernet device: err:0x%x\n", err);
		oskit_netio_release(edev->recv_nio);
		goto bad;
	}
	/*
	 * Release the devs we were not interested in.
	 */
	for (i = 0; i < ndev; i++) {
		if (i != ifnum)
			oskit_etherdev_release(etherdev[i]);
	}

	/*
	 * Stash the edev away.
	 */
	myedevs[ifnum] = edev;
	return 0;

bad:
	for (i = 0; i < ndev; i++) {
		oskit_etherdev_release(etherdev[i]);
	}
	return -1;
}
