/*
 * timer.{cc,hh} -- portable timers. Linux kernel module uses Linux timers
 * Eddie Kohler
 *
 * Copyright (c) 1999 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYRIGHT' file in the source
 * distribution.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "timer.hh"
#include "element.hh"
#include "router.hh"


/*
 * The Timer callback. Call the actual routine.
 */
int
Timer::handler(struct oskit_iunknown *, void *arg)
{
	Timer	*Tp = (Timer *) arg;

	Tp->_hook(Tp->_data);

	return 0;
}

/*
 * Create a Timer based on OSKit clock timers.
 */
Timer::Timer(TimerHook hook, unsigned long thunk)
	: _clock(0), _timer(0), _hook(0), _data(0), _scheduled(0)
{
	oskit_listener_t	*l;
		
	/* Create the clock (only one, and its probably already created). */
	_clock = oskit_clock_init();

	/* Create a timer */
	oskit_clock_createtimer(_clock, &_timer);

	/* Create a listener */
	l = oskit_create_listener(Timer::handler, (void *) this);

	/* Set the listener */
	oskit_timer_setlistener(_timer, l);
	oskit_listener_release(l);

	_hook = hook;
	_data = thunk;
	_scheduled = 0;
}

/*
 * Create a timer for an element.
 */
Timer::Timer(Element *f)
	: _clock(0), _timer(0), _hook(0), _data(0), _scheduled(0)
{
	oskit_listener_t	*l;
		
	/* Create the clock (only one, and its probably already created). */
	_clock = oskit_clock_init();

	/* Create a timer */
	oskit_clock_createtimer(_clock, &_timer);

	/* Create a listener */
	l = oskit_create_listener(Timer::handler, (void *) this);

	/* Set the listener */
	oskit_timer_setlistener(_timer, l);
	oskit_listener_release(l);

	_hook = element_timer;
	_data = (unsigned long) f;
	_scheduled = 0;
}

/*
 * Destroy a Timer.
 */
Timer::~Timer()
{
	if (_timer)
		oskit_timer_release(_timer);
	if (_clock)
		oskit_clock_release(_clock);
}

/*
 * Unschedule a Timer.
 */
void
Timer::unschedule()
{
	if (scheduled()) {
		oskit_itimerspec_t	timespec = { {0, 0}, {0, 0} };

		oskit_timer_settime(_timer, 0, &timespec);
	}
}

/*
 * Schedule a timer.
 */
void
Timer::schedule_after_ms(int ms)
{
	oskit_itimerspec_t	timespec;

	timespec.it_value.tv_sec     = ms / 1000;
	timespec.it_value.tv_nsec    = (ms % 1000) * 1000000;
	timespec.it_interval.tv_sec  = 0;
	timespec.it_interval.tv_nsec = 0;

	_scheduled = 1;
	
	oskit_timer_settime(_timer, 0, &timespec);
}

void
Timer::element_timer(unsigned long thunk)
{
	Element *f = (Element *)thunk;
	f->schedule_tail(); // don't do anything - just put it on the work list
}
